-- HolCrypt.E  X1.01 (C) 2004  Peter Blue / Digital Data Research
-- Experimental Holistic Encryption/Decryption Library

-- This encryption systems works by joining one or more chunks of data into
-- a continious stream and then randomly 'scattering' all the bytes in that
-- stream according to a pre-defined key sequence.
-- The key can be any length.
-- Additional layers of encryption can be applied. A simple XOR system is used here.
-- I have no idea how hard this system would be to crack, perhaps someone can tell me.
-- Functions :-
--   EncryptData(Data,Key)
--   DecryptData(Data,Key)
--   CreateRSB(DataLength,Key) -- Create ReSequence Block
-- [100] 1997-Jul-17 Original concept
-- [101] 2004-Mar-09 Tidy-up. Relesed to RDS

global sequence SystemRSB	SystemRSB = {{},{}} -- Used by block services
global integer  CryptoLayer	CryptoLayer = 3

-- Create ReSequencing Block (RSB)
global function CreateRSB(integer rsl,sequence key)
  integer  idx,idy,kln,kdx
  sequence rsb,rsx
  
  rsx = repeat(0,rsl)
  for i = 1 to rsl do -- Create default sequential index
    rsx[i] = i -- ie 1,2,3,4,5 ...
  end for
  
  kln = length(key)
  if kln > 0 then -- Viable key
    idy = 0 --floor((kln + rsl) / 2)
    rsb = {}
    kdx = 1 -- Index to key values
    while rsl > 0 do
      if kdx > kln then      -- At end of key - Go back to start
        kdx = 1
      end if
      idx = idy + key[kdx]   -- Pick index based on key element
      while idx >= rsl do    -- IDX is greater than RSL
        idx = floor(idx / 2) -- Reduce it until it isn't
      end while              -- (Must be a nicer way to do this)
      idx += 1               -- Eu arrays are 1-based not 0-based
      rsb &= rsx[idx]        -- Add index to RSB list
      rsx[idx] = rsx[rsl]    -- Remove index from RSX list
      rsl -= 1               -- Drop RSL down by 1
      kdx += 1               -- Point to next key element
      idy += 1               -- Assist in scattering process
    end while
  else -- No key - Don't re-sequence
    rsb = rsx
  end if
  
  return rsb -- ie 1,2,3,4,5,6 = 5,3,4,2,1,6
end function

global function SequenceXOR(sequence data,sequence key)
  integer idx,byt,len,kln
  
  idx = 1
  len = length(data)
  kln = length(key)
  for idy = 1 to len do
    byt = data[idy]
    data[idy] = xor_bits(byt,key[idx])
    idx += 1
    if idx > kln then -- At end of key
      idx = 1 -- Go  back to start
    end if
  end for
  
  return data
end function

global function EncryptData(sequence data,sequence key)
  integer len,idx,byt
  sequence out,rsb
  
  len = length(data)
  if and_bits(CryptoLayer,2) = 2 then -- Apply RSB
    out = repeat(0,len) -- Prepare output buffer
    rsb = CreateRSB(len,key)
    for idy = 1 to len do -- Scan along raw data
      idx = rsb[idy]
      byt = data[idx]
      out[idy] = byt
    end for
  else
    out = data
  end if
  
  if and_bits(CryptoLayer,1) = 1 then -- Apply XOR
    out = SequenceXOR(out,key)
  end if
  
  return out
end function

global function DecryptData(sequence data,sequence key)
  integer len,idx,byt
  sequence out,rsb
  
  len = length(data)
  
  if and_bits(CryptoLayer,1) = 1 then -- De-Apply XOR
    data = SequenceXOR(data,key)
  end if
  
  if and_bits(CryptoLayer,2) = 2 then -- De-Apply RSB
    rsb = CreateRSB(len,key)
    out = repeat(0,len)
    for idy = 1 to len do
      idx = rsb[idy]
      byt = data[idy]
      out[idx] = byt
    end for
  else
    out = data
  end if
  
  return out
end function

